from __future__ import annotations

from os import environ
from typing import TYPE_CHECKING, Any

from pytest import fixture, mark

from utilities.platform import IS_NOT_LINUX

if TYPE_CHECKING:
    from collections.abc import Callable

    from sqlalchemy import Engine, Table

FLAKY = mark.flaky(reruns=5, reruns_delay=1)
SKIPIF_CI = mark.skipif("CI" in environ, reason="Skipped for CI")
SKIPIF_CI_AND_NOT_LINUX = mark.skipif(
    ("CI" in environ) and IS_NOT_LINUX, reason="Skipped for CI/non-Linux"
)


# hypothesis

try:
    from utilities.hypothesis import setup_hypothesis_profiles
except ModuleNotFoundError:
    pass
else:
    setup_hypothesis_profiles()


# sqlalchemy


try:
    pass
except ModuleNotFoundError:
    pass
else:

    @fixture(scope="session")
    def create_postgres_engine() -> Callable[..., Engine]:
        """Create a Postgres engine."""

        def inner(*tables_or_mapped_classes: Table | type[Any]) -> Engine:
            from utilities.sqlalchemy import (
                create_engine,
                ensure_tables_created,
                ensure_tables_dropped,
            )

            engine = create_engine(
                "postgresql", host="localhost", port=5432, database="testing"
            )
            ensure_tables_dropped(engine, *tables_or_mapped_classes)
            ensure_tables_created(engine, *tables_or_mapped_classes)
            return engine

        return inner



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import integers

from utilities.weakref import add_finalizer

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture


class TestAddFinalizer:
    @given(n=integers())
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    def test_main(self, *, n: int, capsys: CaptureFixture) -> None:
        @dataclass(frozen=True, kw_only=True)
        class Example:
            n: int

        obj = Example(n=n)

        def callback() -> None:
            print(f"Deleting n={n}...")  # noqa: T201

        out = capsys.readouterr().out
        assert out == ""
        _ = add_finalizer(obj, callback)
        del obj
        out = capsys.readouterr().out
        assert out == f"Deleting n={n}...\n"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar
from weakref import ReferenceType, ref

if TYPE_CHECKING:
    from collections.abc import Callable

_T = TypeVar("_T")


def add_finalizer(obj: _T, callback: Callable[[], None], /) -> ReferenceType[_T]:
    """Add a finalizer for an object."""
    return ref(obj, _add_finalizer_modify(callback))


def _add_finalizer_modify(func: Callable[[], None], /) -> Callable[[Any], None]:
    """Modify a callback to work with `ref`."""

    def wrapped(_: Any, /) -> None:
        func()

    return wrapped


__all__ = ["add_finalizer"]



from __future__ import annotations

from collections.abc import Callable
from functools import cache as _cache
from functools import lru_cache as _lru_cache
from functools import partial as _partial
from typing import Any, ParamSpec, TypeVar, cast, overload

from typing_extensions import override

_T = TypeVar("_T")
_P = ParamSpec("_P")
_R = TypeVar("_R")


def cache(func: Callable[_P, _R], /) -> Callable[_P, _R]:
    """Typed version of `cache`."""
    typed_cache = cast(Callable[[Callable[_P, _R]], Callable[_P, _R]], _cache)
    return typed_cache(func)


@overload
def lru_cache(
    func: Callable[_P, _R], /, *, max_size: int = ..., typed: bool = ...
) -> Callable[_P, _R]: ...
@overload
def lru_cache(
    func: None = None, /, *, max_size: int = ..., typed: bool = ...
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]: ...
def lru_cache(
    func: Callable[_P, _R] | None = None, /, *, max_size: int = 128, typed: bool = False
) -> Callable[_P, _R] | Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Typed version of `lru_cache`."""
    if func is None:
        return cast(
            Callable[[Callable[_P, _R]], Callable[_P, _R]],
            _lru_cache(maxsize=max_size, typed=typed),
        )
    return lru_cache(max_size=max_size, typed=typed)(func)


class partial(_partial[_T]):  # noqa: N801
    """Partial which accepts Ellipsis for positional arguments."""

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> _T:
        iter_args = iter(args)
        head = (next(iter_args) if arg is ... else arg for arg in self.args)
        return self.func(*head, *iter_args, **{**self.keywords, **kwargs})


__all__ = ["cache", "lru_cache", "partial"]
